import { Vector } from "@bearmetal/doodler";
import { SolverTask } from "../../physics/solver.ts";
import { Bogie, Driver } from "./Bogie.ts";
import { TrackSystem } from "../../track/system.ts";
import { clamp } from "../../math/clamp.ts";
import { Train } from "./Train.ts";

export class TrackConstraint extends SolverTask {
  detached: boolean = false;
  detachmentThreshold: number = 1;
  reattachmentThreshold: number = 0.1;
  pathTag?: tag;
  constructor(
    private node: Bogie,
    private track: TrackSystem,
  ) {
    super();
    this.priority = 1;
    this.iterations = 1;
  }
  override apply(): boolean {
    if (this.detached) {
      if (this.node.velocity.mag() < this.reattachmentThreshold) {
        this.detached = false;
      } else {
        this.active = false;
        return false;
      }
    }

    // if (this.node.velocity.mag() < 0) {
    //   return false;
    // }

    const searchRadius = 10;
    const nearbySegments = this.track.grid.query(
      this.node.position,
      searchRadius,
      this.pathTag,
    );

    let closestSegment: GridItem | null = null;
    let closestDistance = Infinity;
    let projectedPoint: Vector | null = null;
    for (const seg of nearbySegments) {
      const candidate = this.closestPointOnSegment(seg, this.node.position);
      const distance = this.node.position.dist(candidate);
      if (distance < closestDistance) {
        closestSegment = seg;
        closestDistance = distance;
        projectedPoint = candidate;
      }
    }

    if (closestSegment && projectedPoint) {
      const correctionForce = Vector.sub(projectedPoint, this.node.position);
      const correctionMag = correctionForce.mag() * this.node.mass;

      if (correctionMag > this.detachmentThreshold) {
        this.detached = true;
        return false;
      }

      this.node.position = projectedPoint;
      const pathDir = Vector.sub(closestSegment[1], closestSegment[0])
        .normalize();
      this.node.velocity.set(pathDir.mult(this.node.velocity.dot(pathDir)));
      if (this.node instanceof Driver) {
        this.node.updateDirection(pathDir);
      }
    }

    return true;
  }

  closestPointOnSegment([a, b]: GridItem, point: Vector): Vector {
    const ab = Vector.sub(b, a);
    const ap = Vector.sub(point, a);
    const t = clamp(ab.dot(ap) / ab.magSq(), 0, 1);

    return Vector.add(a, ab.mult(t));
  }
}

export class MoveTask extends SolverTask {
  constructor(
    private train: Train,
  ) {
    super();
    this.priority = 2;
    this.iterations = 1;
  }

  override apply(dt: number): boolean {
    for (const bogie of this.train.bogies) {
      bogie.update(dt);
    }
    return true;
  }
}

export class TrainTask extends SolverTask {
  constructor(
    private a: Bogie,
    private b: Bogie,
    private restLength: number,
    private tolerance = 0,
    private solid = false,
    private breakThreshold = 10,
  ) {
    super();
    this.iterations = 100;
  }
  override apply(): boolean {
    const delta = Vector.sub(this.b.position, this.a.position);
    const currentLength = delta.mag();
    if (currentLength <= this.tolerance) return false;

    const correction = currentLength - this.restLength;
    const correctionVector = delta.normalize().mult(correction / currentLength);

    const requiredForce = correction / this.restLength;

    if (!this.solid && Math.abs(requiredForce) > this.breakThreshold) {
      this.active = false;
      return false;
    }

    const totalMass = this.a.mass + this.b.mass;
    const aWeight = this.b.mass / totalMass;
    const bWeight = this.b.mass / totalMass;

    this.a.position.add(correctionVector.copy().mult(aWeight));
    this.b.position.sub(correctionVector.copy().mult(bWeight));

    return true;
  }
}
